/* pam_handlers.c -- pam config file parsing and module loading */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dlfcn.h>

#include "pam_appl.h"
#include "pam_private.h"

#define BUF_SIZE      1024
#define MODULE_CHUNK  4

#define DEBUG 0

int _pam_assemble_line(FILE *f, char *buf, int buf_len);

void _pam_free_handlers_aux(struct handler *h);

int _pam_add_handler(pam_handle_t *pamh, int type,
		     int control_flag, char *mod_path,
		     int argc, char **argv);

/* Values for type */
#define PAM_AUTH 0
#define PAM_SESS 1
#define PAM_ACCT 2
#define PAM_PASS 3

/* Parse config file, allocate handler structures, dlopen() */
int _pam_init_handlers(pam_handle_t *pamh)
{
    FILE *f;
    char buf[BUF_SIZE];
    char *tok, *mod_path;
    int x, module_type, control_flag;
    char *service_name = pamh->service_name;

    int argc, l;
    char *argvbuf;
    char *argvbufp;
    char **argv;

    /* First clean the service structure */
    
    if (! pamh->handlers.module) {
	pamh->handlers.module = malloc(MODULE_CHUNK *
				       sizeof(struct loaded_module));
	pamh->handlers.modules_allocated = MODULE_CHUNK;
	pamh->handlers.modules_used = 0;
    }
    pamh->handlers.authenticate = NULL;
    pamh->handlers.setcred = NULL;
    pamh->handlers.acct_mgmt = NULL;
    pamh->handlers.open_session = NULL;
    pamh->handlers.close_session = NULL;
    pamh->handlers.chauthtok = NULL;

    /* Now parse the config file and add handlers */
    
    f = fopen(PAM_CONFIG, "r");
    while ((x = _pam_assemble_line(f, buf, BUF_SIZE)) > 0) {
	tok = strtok(buf, " \n\t");
	if (!strcmp(tok, service_name)) {
	    /* This is the service we are looking for */
	    /* XXX - need to handle OTHER */

	    tok = strtok(NULL, " \n\t");
	    if (!strcmp("auth", tok)) {
		module_type = PAM_AUTH;
	    } else if (!strcmp("session", tok)) {
		module_type = PAM_SESS;
	    } else if (!strcmp("account", tok)) {
		module_type = PAM_ACCT;
	    } else if (!strcmp("password", tok)) {
		module_type = PAM_PASS;
	    } else {
		/* XXX - Illegal module type */
		return -1;
	    }

	    tok = strtok(NULL, " \n\t");
	    if (!strcmp("required", tok)) {
		control_flag = PAM_REQUIRED;
	    } else if (!strcmp("optional", tok)) {
		control_flag = PAM_OPTIONAL;
	    } else if (!strcmp("sufficient", tok)) {
		control_flag = PAM_SUFFICIENT;
	    } else {
	        /* XXX - Illegal flag */
	        return -1;
	    }

	    tok = strtok(NULL, " \n\t");
	    mod_path = tok;

	    tok += strlen(tok) + 1;
	    /* Overkill on the malloc, but not large */
	    argc = 0;
	    l = strlen(tok);
	    argv = NULL;
	    if (l) {
		argvbuf = malloc(l * (sizeof(char) + sizeof(char *)));
		argv = argvbuf;
		argvbufp = argvbuf + l * sizeof(char *);
		while ((tok = strtok(tok, " \n\t"))) {
		    strcpy(argvbufp, tok);
		    *argv = argvbufp;
		    argvbufp += strlen(argvbufp) + 1;
		    
		    argc++;
		    argv++;
		    tok = NULL;
		}
		argv = argvbuf;
	    }
#if DEBUG
	    printf("CONF: %s %d %d %s %d\n", service_name, module_type,
		   control_flag, mod_path, argc);
	    for (l = 0; l < argc; l++) {
		printf("CONF: %s\n", argv[l]);
	    }
#endif
	    _pam_add_handler(pamh, module_type, control_flag,
			     mod_path, argc, argv);
	}
    }
    fclose(f);

    if (x < 0) {
	/* XXX - Read error */
	return -1;
    }

    return PAM_SUCCESS;
}

int _pam_assemble_line(FILE *f, char *buffer, int buf_len)
{
    char *p = buffer;
    char *s, *os;
    int used = 0;

    while (1) {
	if (used >= buf_len) {
	    /* Overflow */
	    return -1;
	}
	if (!fgets(p, buf_len - used, f)) {
	    if (used) {
		/* Incomplete read */
		return -1;
	    } else {
		/* EOF */
		return 0;
	    }
	}
	s = p + strspn(p, " \n\t");
	if (*s && (*s != '#')) {
	    /* Check for backslash */
	    os = s;
	    s += strlen(s);
	    while ((*s == ' ') || (*s == '\t')) {
		s--;
	    }
	    if (*s == '\\') {
		*s++ = ' ';
		*s = '\0';
		used += strlen(os);
		p = s;
	    } else {
		/* End of the line! */
		used += strlen(os);
		break;
	    }
	} else {
	    /* Nothing in this line */
	    /* Don't move p         */
	}
    }

    return used;
}

/* XXX - _pam_add_handler() is where we would handle the static libs */

int _pam_add_handler(pam_handle_t *pamh, int type,
		     int control_flag, char *mod_path,
		     int argc, char **argv)
{
    struct loaded_module *mod = pamh->handlers.module;
    int x = 0;
    struct handler **handler_p;
    struct handler **handler_p2;
    int (*func)();
    int (*func2)();
    
    /* First, ensure the module is loaded */
    while (x < pamh->handlers.modules_used) {
	if (!strcmp(mod[x].name, mod_path)) {
	    break;
	}
	x++;
    }
    if (x == pamh->handlers.modules_used) {
	/* Not found */
	if (pamh->handlers.modules_allocated == pamh->handlers.modules_used) {
	    pamh->handlers.modules_allocated += MODULE_CHUNK;
	    pamh->handlers.module = realloc(pamh->handlers.module,
					    pamh->handlers.modules_allocated);
	}
	mod = &(pamh->handlers.module[x]);
	mod->name = strdup(mod_path);
	/* XXX check errors */
	mod->dl_handle = dlopen(mod_path, RTLD_NOW);
	if (! mod->dl_handle) {
	    printf("Could not load %s\n", mod_path);
	}
    }

    /* mod points to the struct loaded_module we care about */
    /* Now add the handler based on mod->dlhandle and type  */

    handler_p2 = NULL;
    func2 = NULL;
    switch (type) {
    case PAM_AUTH:
	handler_p = &pamh->handlers.authenticate;
	func = dlsym(mod->dl_handle, "pam_sm_authenticate");
	break;
    case PAM_SESS:
	handler_p = &pamh->handlers.open_session;
	func = dlsym(mod->dl_handle, "pam_sm_open_session");
	handler_p2 = &pamh->handlers.close_session;
	func2 = dlsym(mod->dl_handle, "pam_sm_close_session");
	break;
    case PAM_ACCT:
	handler_p = &pamh->handlers.setcred;
	func = dlsym(mod->dl_handle, "pam_sm_setcred");
	handler_p2 = &pamh->handlers.acct_mgmt;
	func2 = dlsym(mod->dl_handle, "pam_sm_acct_mgmt");
	break;
    case PAM_PASS:
	handler_p = &pamh->handlers.chauthtok;
	func = dlsym(mod->dl_handle, "pam_sm_chauthtok");
	break;
    default:
	/* XXX - Illegal module type */
	return -1;
    }

    while (*handler_p) {
	handler_p = &((*handler_p)->next);
    }
    *handler_p = malloc(sizeof(struct handler));
    (*handler_p)->func = func;
    (*handler_p)->control_flag = control_flag;
    (*handler_p)->argc = argc;
    (*handler_p)->argv = argv;
    (*handler_p)->next = NULL;

    if (handler_p2) {
	while (*handler_p2) {
	    handler_p2 = &((*handler_p2)->next);
	}
	*handler_p2 = malloc(sizeof(struct handler));
	(*handler_p2)->func = func2;
	(*handler_p2)->control_flag = control_flag;
	(*handler_p2)->argc = argc;
	(*handler_p2)->argv = argv;
	(*handler_p2)->next = NULL;
    }
    
    return PAM_SUCCESS;
}

/* Free various allocated structures and dlclose() the libs */
int _pam_free_handlers(pam_handle_t *pamh)
{
    struct loaded_module *mod = pamh->handlers.module;

    /* Close all loaded modules */
    while (pamh->handlers.modules_used--) {
	free(mod->name);
	dlclose(mod->dl_handle);
	mod++;
    }

    _pam_free_handlers_aux(pamh->handlers.authenticate);
    _pam_free_handlers_aux(pamh->handlers.setcred);
    _pam_free_handlers_aux(pamh->handlers.acct_mgmt);
    _pam_free_handlers_aux(pamh->handlers.open_session);
    _pam_free_handlers_aux(pamh->handlers.close_session);
    _pam_free_handlers_aux(pamh->handlers.chauthtok);

    pamh->handlers.authenticate = NULL;
    pamh->handlers.setcred = NULL;
    pamh->handlers.acct_mgmt = NULL;
    pamh->handlers.open_session = NULL;
    pamh->handlers.close_session = NULL;
    pamh->handlers.chauthtok = NULL;
    
    return PAM_SUCCESS;
}

void _pam_free_handlers_aux(struct handler *h)
{
    struct handler *last;

    while (h) {
	last = h;
	free(h->argv);  /* This is all alocated in a single chunk */
	h = h->next;
	free(last);
    }
}
